indexing
	description: "Lists have a head an a tail. Lists are composed of cells."

class IMMUTABLE_LIST [G]

inherit
	ANY
		redefine
			is_equal
		end

create
	make

feature {NONE} -- representation

	first_cell: IMMUTABLE_CELL [G]
			-- The list is a chain of immutable cells

	secret_count : INTEGER
			-- Internal count (the visible 'count' feature is
			-- advertised as derived from 'is_empty')

feature {IMMUTABLE_LIST} -- support for making new lists

	set_first_cell( a_cell : IMMUTABLE_CELL[G]; a_count : INTEGER ) is
			-- Initialize 'Current' to hold the list of cells beginning
			-- with 'a_cell' (which can be Void).
			--
			-- 'a_count' should be the number of cells in the list of
			-- cells beginning with 'a_cell'.
		do
			first_cell := a_cell
			secret_count := a_count
		ensure
			first_cell_set: first_cell = a_cell
			secret_count_set: secret_count = a_count
		end

feature -- creation

	make is
			-- Make a new empty list
		do
			first_cell := Void
		ensure
			empty: is_empty
		end

feature -- basic queries

	is_empty : BOOLEAN is
			-- Does the list contain no items?
		do
			Result := ( secret_count = 0 )
		end

	head:G is
			-- The first element in the list
		require
			not_empty: not is_empty
		do
			Result := first_cell.item
		end

	tail:IMMUTABLE_LIST [G] is
			-- A new list formed from 'Current' minus its 'head'
		require
			not_empty: not is_empty
		do
			create Result.make
			Result.set_first_cell(first_cell.right, secret_count - 1 )
		end

feature -- derived queries

	count : INTEGER is
			-- The number of items in the list
		do
			Result := secret_count
		ensure
			is_empty_means_count_is_zero:
				is_empty implies (Result = 0)
			else_count_is_one_greater_than_count_of_tail:
				(not is_empty) implies (Result = 1 + tail.count)
		end

	preceded_by (g:G) : IMMUTABLE_LIST [G] is
			-- A new list formed from Current with 'g' added at the head.
		local
			new_cell : IMMUTABLE_CELL[ G ]
		do
			create new_cell.make( g, first_cell )
			create Result.make
			Result.set_first_cell(new_cell, secret_count + 1)
		ensure
			head_is_g: Result.head  = g
			tail_is_original_list: Result.tail.is_equal( Current )
			not_empty: not Result.is_empty
		end

	is_equal ( other: IMMUTABLE_LIST [G]): BOOLEAN is
			-- Does 'Current' contain the same items as
			-- 'other', in the same order?
		require else
			other_exists:
				other /= Void
		do
			if
				-- Case 1: ' Current' and 'other' are both empty
				(secret_count = 0) and (other.count = 0)
			then
				Result := true
			else
				if
					-- Case 2: only one of the lists is empty
					(secret_count = 0) or (other.count = 0)
				then
					Result := false
				else
					-- Case 3: neither list is empty
					if
						-- Case 3.1: the heads of the two lists are not equal
						head /= other.head
					then
						Result := false
					else
						-- Case 3.2: the two heads are equal; result depends upon tails
						Result := tail.is_equal (other.tail)
					end
				end
			end
		ensure then
			same_contents:
				Result =
						-- both lists are empty
						( (count = 0) and (other.count = 0) )
					or else
						-- neither list is empty, and ...
						(
								( (count > 0) and (other.count > 0) )
							and then
								-- ... their heads are equal and their tails are equal.
								(head = other.head and tail.is_equal (other.tail) )
						)
		end

	item( i : INTEGER ) : G is
			-- The i-th item in the list (the head is item(1) )
		require
			i_large_enough: i >= 1
			i_small_enough: i <= count
		do
			if
				i = 1
			then
				Result := head
			else
				Result := tail.item( i-1 )
			end
		ensure
				( i = 1 and Result = head )
			or else
				Result = tail.item( i - 1 )
		end

	sublist( from_position : INTEGER; to_position : INTEGER ) : IMMUTABLE_LIST [G] is
			-- A new list formed from the items
			-- at 'from_position' through 'to_position'
		require
			from_position_large_enough: from_position >= 1
			from_position_small_enough: from_position <= to_position + 1
			to_position_small_enough: to_position <= count
		local
			index : INTEGER
			current_cell,		-- pointer to the cell in Current we are copying from
			result_list,		-- the cells that form the result
			result_last,		-- pointer to the last cell in the result list
			new_cell : IMMUTABLE_CELL[ G ]
		do
			-- Make a new, empty result list
			create Result.make
			if
				-- the result should have one or more items in it
				from_position <= to_position
			then
				-- Work along 'Current' until we reach the 'from_position'
				from
					index := 1
					current_cell := first_cell
				until
					index = from_position
				loop
					current_cell := current_cell.right
					index := index + 1
				end

				-- Start the result list with a new cell containing the
				-- item at the 'from_position' (there must always be
				-- at least one item to copy - sublist cannot be used
				-- to create an empty list). This first cell in the
				-- result list is also (for the moment) the last.
				create new_cell.make( current_cell.item, Void )
				result_list := new_cell
				result_last := new_cell

				-- Work along 'Current' to the 'to_position', making new
				-- cells to hold items from 'Current' and appending them
				-- to the end of the result list
				from
					index := index + 1
					current_cell := current_cell.right
				until
					index > to_position
				loop
					create new_cell.make( current_cell.item, Void )
					result_last.set_right( new_cell )
					result_last := new_cell
					index := index + 1
					current_cell := current_cell.right
				end

				-- Finally, set the cells in Result to this 'result_list'
				Result.set_first_cell( result_list, (to_position - from_position + 1 ) )
			end
		ensure
			is_empty_consistent_with_from_and_to_position:
				Result.is_empty = ( from_position > to_position )
			Result_head_is_at_from_position_in_Current:
				(from_position <= to_position) implies
					( Result.head = item( from_position ) )
			Result_tail_is_correct_sublist_within_Current:
				(from_position < to_position) implies
					( Result.tail.is_equal( sublist( from_position + 1, to_position ) ) )
		end

end
