indexing
	description: "Simple queue with contracts written in terms of immutable lists."

class SIMPLE_QUEUE[ G ]

create
	make

feature {NONE} -- representation (queue elements are stored in an array, which is treated as if it were circular)

	contents : ARRAY[ G ]
	position_of_first : INTEGER
	position_of_last : INTEGER
	-- The public feature 'count' is also part of the representation

	next_location_after( i : INTEGER ) : INTEGER is
			-- The next queue location after location i, treating the
			-- array as circular (the array has locations 1 through 'capacity').
		do
			Result := (i \\ capacity) + 1
		ensure
			if_not_at_end_then_its_a_plain_increment: (i < capacity) implies (Result = i + 1)
			if_at_end_then_its_back_to_one: (i = capacity) implies (Result = 1)
		end


feature -- 1. basic queries

	capacity : INTEGER
		-- Maximum number of items

	items : IMMUTABLE_LIST[ G ] is
			-- The items in the queue, in queue order
		local
			index_into_contents : INTEGER
			temp : IMMUTABLE_LIST[ G ]
		do
			from
				create temp.make
				index_into_contents := position_of_first
			until
				temp.count = count
			loop
				temp := temp.preceded_by( contents.item(index_into_contents) )
				index_into_contents := next_location_after( index_into_contents )
			end
			from
				create Result.make
			until
				temp.count = 0
			loop
				Result := Result.preceded_by( temp.head )
				temp := temp.tail
			end
		end

feature -- 2. derived queries

	count : INTEGER
			-- The number of items in the queue
		-- ensure
		-- 	Result = items.count
		-- 		(See invariant.)

	is_empty : BOOLEAN is
			-- Does the queue contain no items?
		do
			Result := ( count = 0 )
		ensure
			consistent_with_items_count: Result = ( items.count = 0 )
		end

	is_full : BOOLEAN is
			-- Does the queue have no room for another item?
		do
			Result := ( count = capacity )
		ensure
			consistent_with_items_count_and_capacity: Result = ( items.count = capacity )
		end

	head : G is
			-- The first item in the queue
		do
			Result := contents.item( position_of_first )
		ensure
			consistent_with_items: Result = items.head
		end

feature -- 3. creation commands


	make( a_capacity : INTEGER ) is
			-- Make an empty queue with the given capacity
		require
			capacity_large_enough: a_capacity >= 1
		do
			create contents.make( 1, a_capacity )
			capacity := a_capacity
			count := 0

			--- After the very first call to 'put' we want 'position_of_first'
			-- and 'position_of_last' both to be pointing to location the same location
			-- (and, arbitrarily, we choose it to be location 1).
			--
			-- The call to 'put' will move 'position_of_last' on one, so initially
			-- we need 'position_of_first' to be 1 and 'position_of_last' to be 'capacity'
			position_of_first := 1
			position_of_last := capacity
		ensure
			capacity_set: capacity = a_capacity
			empty: items.count = 0
			representation_set: (position_of_first = 1) and (position_of_last = capacity)
		end

feature -- 4. other commands

	put( g : G ) is
			-- Add 'g' to the queue as the oldest item
		require
			not_full: count < capacity
		do
			position_of_last := next_location_after( position_of_last )
			contents.put( g, position_of_last )
			count := count + 1
		ensure
			count_increased: items.count = old items.count + 1
			g_is_the_youngest_item: items.item(count) = g
		end


	remove is
			-- Remove the oldest item from the queue
		require
			not_empty: count > 0
		do
			count := count - 1
			position_of_first := next_location_after( position_of_first )
		ensure
			count_decreased: items.count = old items.count - 1
			items_left_shifted: items.is_equal( old items.tail )
				-- informally:
				--   forall i in the range 1 to the count of the current items,
				--     items @ i = old ( items @ (i+1) )
		end


invariant
	-- Regular
		capacity_is_at_least_one: capacity >= 1
		count_never_negative: items.count >= 0
		count_never_greater_than_capacity: items.count <= capacity
		-- Constraints on queries
			count_consistent_with_items: count = items.count
end
