indexing
	description: "Simple queue with contracts that cover frame rules";

class SIMPLE_QUEUE_WITH_FRAME_RULES[ G ]

creation
	make

feature {NONE} -- bug control

	put_bug_1 : BOOLEAN is false	-- violates 'original_items_remain' in 'put'
	put_bug_2 : BOOLEAN is true	-- violates 'original_items_unchanged' in 'put' 

feature {NONE} -- representation (queue elements are stored in an array, which is treated as if it were circular)

	contents : ARRAY[ G ]
	position_of_first : INTEGER
	position_of_last : INTEGER
	number_of_items : INTEGER
	-- The value of this attribute is presented to clients as the derived query 'count'.
	
	next_location_after( i : INTEGER ) : INTEGER is
			-- The next queue location after location i, treating the
			-- array as circular (the array has locations 1 through 'capacity').
		do
			Result := (i \\ capacity) + 1
		ensure
			if_not_at_end_then_its_a_plain_increment: (i < capacity) implies (Result = i + 1)
			if_at_end_then_its_back_to_one: (i = capacity) implies (Result = 1)
		end


feature -- 1. basic queries

	capacity : INTEGER
		-- Maximum number of items

	items : IMMUTABLE_LIST[ G ] is
			-- The items in the queue, in queue order
		local
			index_into_contents : INTEGER
			temp : IMMUTABLE_LIST[ G ]
		do
			from
				create temp.make
				index_into_contents := position_of_first
			until
				temp.count = number_of_items
			loop
				temp := temp.preceded_by( contents.item(index_into_contents) )
				index_into_contents := next_location_after( index_into_contents )
			end
			from
				create Result.make
			until
				temp.count = 0
			loop
				Result := Result.preceded_by( temp.head )
				temp := temp.tail
			end
		end

feature -- 2. derived queries

	count : INTEGER is
			-- The number of items in the queue
		do
			Result := number_of_items
		ensure
			consistent_with_items_count:
				Result = items.count
		end

	is_empty : BOOLEAN is
			-- Does the queue contain no items?
		do
			Result := ( number_of_items = 0 )
		ensure
			consistent_with_items_count: Result = ( items.count = 0 )
		end

	is_full : BOOLEAN is
			-- Does the queue have no room for another item?
		do
			Result := ( number_of_items = capacity )
		ensure
			consistent_with_items_count_and_capacity: Result = ( items.count = capacity )
		end

	first : G is	
			-- The oldest item in the queue
		do
			Result := contents.item( position_of_first )
		ensure
			consistent_with_items: Result = items.head
		end

feature -- 3. creation commands

	
	make( a_capacity : INTEGER ) is
			-- Make an empty queue with the given capacity
		require
			capacity_large_enough: a_capacity >= 1
		do
			create contents.make( 1, a_capacity )
			capacity := a_capacity
			number_of_items := 0

			--- After the very first call to 'put' we want 'position_of_first' 
			-- and 'position_of_last' both to be pointing to the same location
			-- (and, arbitrarily, we choose it to be location 1).
			--
			-- The call to 'put' will move 'position_of_last' on one, so initially
			-- we need 'position_of_first' to be 1 and 'position_of_last' to be 'capacity'
			position_of_first := 1
			position_of_last := capacity
		ensure
			capacity_set: capacity = a_capacity
			empty: items.count = 0
			representation_set: (position_of_first = 1) and (position_of_last = capacity)
		end

feature -- 4. other commands

	put( g : G ) is
			-- Add 'g' to the queue as the oldest item
		require
			not_full: count < capacity
		do
			position_of_last := next_location_after( position_of_last )
			contents.put( g, position_of_last )
			number_of_items := number_of_items + 1
			if 
				put_bug_1
			then
				-- copy pointer to item 2 into location 1
				if
					count >= 2
				then
					contents.put( contents.item(2), 1 )
				end
			end
			if 
				put_bug_2
			then
				-- copy what's in the item in location 2 into the item in location 1
				if
					count >= 2
				then
					contents.item(1).copy( contents.item(2) )
				end
			end
		ensure
			number_of_items_increased: items.count = old items.count + 1 
			g_is_the_youngest_item: items.item(items.count) = g
			capacity_unchanged: capacity = old capacity
			original_items_remain:
				( items.count>=2 ) implies (items.sublist(1, items.count-1).is_equal( old items ) )
			original_items_unchanged: 
				items.count >= 2 implies 
					( deep_equal( items.sublist(1, items.count-1), old deep_clone(items) ) )
		end


	remove is
			-- Remove the oldest item from the queue
		require
			not_empty: count > 0
		do
			number_of_items := number_of_items - 1
			position_of_first := next_location_after( position_of_first )
		ensure
			number_of_items_decreased: items.count = old items.count - 1
			items_left_shifted: items.is_equal( old items.tail )
				-- informally:
				--   forall i in the range 1 to the count of the current items,
				--     items @ i = old ( items @ (i+1) )
		end


invariant

	capacity_is_at_least_one: capacity >= 1
	count_never_negative: items.count >= 0
	count_never_greater_than_capacity: items.count <= capacity
end
