class SIMPLE_QUEUE_WITH_FRAME_RULES[ G ]

creation
	make

feature {NONE} -- bug control

	bug_in_put_routine : BOOLEAN is false
	bug_in_remove_routine : BOOLEAN is false

feature {NONE} -- representation (queue elements are stored in an array, which is treated as if it were circular)

	contents : ARRAY[ G ]
	position_of_first : INTEGER
	position_of_last : INTEGER
	
	next_location_after( i : INTEGER ) : INTEGER is
			-- The next queue location after location i, treating the
			-- array as circular (the array has locations 1 through 'capacity').
		do
			Result := (i \\ capacity) + 1
		ensure
			if_not_at_end_then_go_to_next: (i < capacity) implies (Result = i + 1)
			if_at_end_then_restart_at_one: (i = capacity) implies (Result = 1)
		end


feature -- 1. basic queries

	capacity : INTEGER
			-- Maximum number of items

	count : INTEGER
			-- Number of items in queue

	item( i : INTEGER ) : G is
			-- The i-th item in the queue, 
			-- where item(1) is the oldest item.
		require
			i_big_enough: i >= 1
			i_small_enough: i <= count
		local
			index : INTEGER
				-- The location in the circular array
				-- corresponding to 'i'
		do
			index := (position_of_first + i - 1) \\ capacity
			if
				index = 0
			then
				index := capacity
			end
			Result := contents.item( index )
		end

feature -- 2. derived queries

	is_empty : BOOLEAN is
			-- Does the queue contain no items?
		do
			Result := ( count = 0 )
		ensure
			consistent_with_count: Result = ( count = 0 )
		end

	is_full : BOOLEAN is
			-- Does the queue have no room for another item?
		do
			Result := ( count = capacity )
		ensure
			consistent_with_count_and_capacity: Result = ( count = capacity )
		end

	first : G is	
			-- The oldest item in the queue
		do
			Result := contents.item( position_of_first )
		ensure
			consistent_with_item: Result = item( 1 )
		end

feature -- 3. creation commands

	
	make( a_capacity : INTEGER ) is
			-- Make an empty queue with the given capacity
		require
			capacity_large_enough: a_capacity >= 1
		do
			!!contents.make( 1, a_capacity )
			capacity := a_capacity
			count := 0

			-- After the very first call to 'put' we want 'position_of_first' 
			-- and 'position_of_last' both to be pointing to the same location
			-- (and, arbitrarily, we choose it to be location 1).
			---
			-- The call to 'put' will move 'position_of_last' on one, so initially
			-- we need 'position_of_first' to be 1 and 'position_of_last' to be 'capacity'
			position_of_first := 1
			position_of_last := capacity
		ensure
			capacity_set: capacity = a_capacity
			empty: count = 0
			representation_set: (position_of_first = 1) and (position_of_last = capacity)
		end

feature -- 4. other commands

	put( g : G ) is
			-- Add 'g' to the queue as the oldest item
		require
			not_full: count < capacity
		do
			put__for_all_old_1_save ( g,  1 ,   count) 
			position_of_last := next_location_after( position_of_last )
			contents.put( g, position_of_last )
			count := count + 1
			if bug_in_put_routine and then count >= 2 then contents.put( contents.item(2), 1 ) end
		ensure
			number_of_items_increased: count = old count + 1
			g_is_the_youngest_item: item(count) = g
			original_items_unchanged: -- for_all i:INTEGER in 1 .. old count, item(i) = old item(i)
			  put__for_all_i ( g,  1 ,  old count)
		end


	remove is
			-- Remove the oldest item from the queue
		require
			not_empty: count > 0
		do
			remove__for_all_old_1_save ( 2 ,   count) 
			count := count - 1
			position_of_first := next_location_after( position_of_first )
			if bug_in_remove_routine then contents.put( contents.item(3), 2 ) end
		ensure
			number_of_items_decreased: count = old count - 1
			items_left_shifted: -- for_all i:INTEGER in 2 .. old count, item(i-1) = old item(i)
			  remove__for_all_i ( 2 ,  old count)
		end




feature {NONE}
	-- implementation of universal and existential quantifiers

	put__for_all_old_1: INDEXED_COLLECTION [like  item]

	put__for_all_old_1_save( g : G; for_all_i_from, for_all_i_until: INTEGER) is 
		local
			i:INTEGER
		do
			from
				i := for_all_i_from
				!!put__for_all_old_1.make(for_all_i_from, for_all_i_until)
			until
				i > for_all_i_until
			loop
				put__for_all_old_1.put( item(i),i )
				i := i + 1
			end
		end

	put__for_all_i( g : G; for_all_i_from, for_all_i_until:INTEGER): BOOLEAN is
		local
			i:INTEGER
		do
			from
				Result := true
				i := for_all_i_from
			until
				i > for_all_i_until or not Result
			loop
				Result :=  item(i) =put__for_all_old_1.item(i ) 
				i := i + 1
			end
		end

	remove__for_all_old_1: INDEXED_COLLECTION [like  item]

	remove__for_all_old_1_save(for_all_i_from, for_all_i_until: INTEGER) is 
		local
			i:INTEGER
		do
			from
				i := for_all_i_from
				!!remove__for_all_old_1.make(for_all_i_from, for_all_i_until)
			until
				i > for_all_i_until
			loop
				remove__for_all_old_1.put( item(i),i )
				i := i + 1
			end
		end

	remove__for_all_i(for_all_i_from, for_all_i_until:INTEGER): BOOLEAN is
		local
			i:INTEGER
		do
			from
				Result := true
				i := for_all_i_from
			until
				i > for_all_i_until or not Result
			loop
				Result :=  item(i-1) =remove__for_all_old_1.item(i ) 
				i := i + 1
			end
		end
invariant

	capacity_at_least_one: capacity >= 1
	count_never_negative: count >= 0
	count_never_greater_than_capacity: count <= capacity
end