indexing
	description:
		" Objects of subclasses of ABSTRACT_OBSERVER register with objects %
		% of subclasses of ABSTRACT_SUBJECT if they want to be notified %
		% that their subject has changed."

class ABSTRACT_SUBJECT

inherit
	SUBJECT_TYPE_WITH_FRAMING_AND_VISIBILITY_CONTROL
		export {NONE} observers end
	BUG_SWITCHES

create
	initialize

feature {NONE} -- representation

	the_attached_observers : ARRAY[ ABSTRACT_OBSERVER ]

	count : INTEGER
		-- The number of attached observers


feature -- creation

	initialize is
			-- Initialise this subject to have no observers
		do
			create the_attached_observers.make(1, 100)
					-- cheap implementation for now (goal is to test contracts)
		end



feature {NONE} -- basic query

	observers : IMMUTABLE_SET[ ABSTRACT_OBSERVER ] is
			-- A newly-created set containing the
			-- observers attached to this subject
		local
			i : INTEGER
		do
			create Result.initialize
			if
				the_attached_observers /= Void
			then
				from
					i := 1
				until
					i > count
				loop
					Result := Result.plus(
						the_attached_observers.item(i) )
					i := i + 1
				end
			end
		end


feature -- attach and detach observers

	attached( o : ABSTRACT_OBSERVER ) : BOOLEAN is
			-- Is 'o' one of this subject's
			-- attached observers?
		local
			i : INTEGER
		do
			-- Implementation note: can't simply use
			-- 'Result := the_attached_observers.has( o )'
			-- because an observer 'o' might be in the array
			-- but not in the part indexed by 1..count.
			from
				i := 1
			until
				i > count
			or
				Result
			loop
				Result := the_attached_observers.item(i) = o
				i := i + 1
			end
		end


	attach( o : ABSTRACT_OBSERVER ) is
			-- Add 'o' to the set of objects that
			-- will be informed if this subject changes.
		local
			i : INTEGER --- for bug planting
			existing_observer : ABSTRACT_OBSERVER --- for bug planting
		do
			if ---
				bug__attach_loses_observer_i > 0 ---
			then ---
				if ---
					bug__attach_loses_observer_i <= count ---
				then ---
					detach( the_attached_observers.item(bug__attach_loses_observer_i) ) ---
				end ---
			end ---

			if --
				not bug__observer_not_attached --
			then --
				count := count + 1
				the_attached_observers.put( o, count )
			end --

			if ---
				bug__attach_duplicates_an_existing_observer
			then ---
				if ---
					the_attached_observers.count > 0 ---
				then ---
					existing_observer := the_attached_observers.item(1) ---
					count := count + 1
					the_attached_observers.put( existing_observer, count ) ---
				end ---
			end ---
		end

	detach( o : ABSTRACT_OBSERVER ) is
			-- Remove 'o' from the set of objects that
			-- will be informed if this subject changes.
		local
			position_of_o : INTEGER
		do
			-- For the precondition to be true, at least
			-- one observer must have been attached. As
			-- a consequence, we can be sure that
			-- 'the_attached_observers /= Void'

			from
				position_of_o := 1
			until
				o = the_attached_observers.item( position_of_o )
			loop
				position_of_o := position_of_o + 1
			end
			-- Overwrite the observer to be detached with the last
			-- in the array, and decrement the count
			the_attached_observers.put(
				the_attached_observers.item(count), position_of_o )
			count := count - 1
		end

feature -- send notifications

	notify is
			-- Inform attached observers that this subject has
			-- changed by calling 'abstract_update' on each one.
			--
			-- 'notify' should be called by any
			-- subclass feature that changes the state of
			-- 'Current' in a way that its observers should
			-- know about.
		local
			old_attached_observers : ARRAY[ ABSTRACT_OBSERVER ]
			old_observer_count : INTEGER
			i : INTEGER
		do
			if ---
				not bug__subject_fails_to_respond_to_notify ---
			then ---
				if
					the_attached_observers /= Void
				then
					from
						i := 1
					until
						i > count
					loop
						if ---
							bug__subject_fails_to_inform_observer_i /= i ---
						then ---
							if
								the_attached_observers.item(i) /= Void
							then
								the_attached_observers.item(i).update
							end
						end ---
						i := i + 1
					end
				end
			end ---
		end

	all_observers_up_to_date( current_observers : like observers ) : BOOLEAN is
		do
			Result :=
				(
					( current_observers.count = 0 )
				  or else
					(
						current_observers.choice.up_to_date_with_subject
					  and
						all_observers_up_to_date(
								current_observers.minus( current_observers.choice ) )
					)
				)
	end



end -- class ABSTRACT_SUBJECT

