indexing
	description:
		"Immutable set objects (contracts omitted)"

class IMMUTABLE_SET[ G ]

inherit
	ANY
		redefine
			copy,
			is_equal
		end

create
	initialize

feature {NONE} -- representation

	the_contents : ARRAY[ G ]

feature {IMMUTABLE_SET} -- set construction

	put( g : G ) is
			-- A command for use in constructing
			-- new sets that form the results of
			-- the public functions.
		do
			count := count + 1
			the_contents.put( g, count )
		end

feature --{IMMUTABLE_SET} -- low-level access

	get( i : INTEGER ) : G is
			-- The i-th element in the representation
			-- of the set (protects public features from
			-- knowing full details of representation)
		do
			Result := the_contents.item(i)
		end

feature -- initialization

	initialize is
			-- Initialize set to be empty
		do
			create the_contents.make( 1, 100 )
		end

feature


	count : INTEGER
			-- The number of elements in the set

	is_empty : BOOLEAN is
			-- Does the set contain no elements?
		do
			Result := ( count = 0 )
		end

	has( g : G ) : BOOLEAN is
			-- Is 'g' an element of 'Current'?
		local
			i : INTEGER
		do
			from
				i := 1
			until
				i > count
			or
				Result
			loop
				Result := g = get(i)
				i := i + 1
			end
		end

	is_subset( other : like Current ) : BOOLEAN is
			-- Are all the elements in 'Current' also
			-- in 'other'?
		local
			i : INTEGER
		do
			from
				i := 1
				Result := ( count <= other.count )
			until
				i > count
			or
				not Result
			loop
				Result := other.has( get(i) )
				i := i + 1
			end
		end

	is_equal( other : like Current ) : BOOLEAN is
			-- Does 'other' contain exactly the same elements
			-- as 'Current'?
		do
			Result :=
					is_subset( other )
				and
					other.is_subset( Current )
		end

	plus( g : G ) : like Current is
			-- A new set containing all the elements
			-- of 'Current' and 'g', too.
		require
			non_void_argument: g /= Void
		local
			i : INTEGER
		do
			from
				i := 1
				create Result.initialize
			until
				i > count
			loop
				Result.put( get(i) )
				i := i + 1
			end
			if
				not has( g )
			then
				Result.put( g )
			end
		end

	choice : G is
			-- An arbitrarily (but deterministically) chosen
			-- element.
		require
			not_empty: count > 0
		do
			Result := get( 1 )
		end

	minus( g : G ) : like Current is
			-- A new set containing all the elements of
			-- 'Current' except 'g' (in the special case
			-- that 'Current' does not contain 'g', 'minus'
			-- yields a clone of 'Current')
		local
			i : INTEGER
		do
			from
				i := 1
				create Result.initialize
			until
				i > count
			loop
				if
					get(i) /= g
				then
					Result.put( get(i) )
				end
				i := i + 1
			end
		end

feature {NONE} -- by redefining 'copy', we give 'clone' its proper behaviour

	copy (other: like Current) is
			-- Update current object using fields of object attached
			-- to 'other', so as to yield equal objects.
			-- (NOTE: sometime, rewrite not using 'get')
		local
			i : INTEGER
		do
			from
				i := 1
				count := 0
				create the_contents.make(1, 100)
			until
				i > other.count
			loop
				put( other.get(i) )
				i := i + 1
			end
		end
end
