indexing
	description:
		" Observers register with subjects if they want to be informed %
		% that their subject has changed. %
		% %
		% SUBJECT_TYPE defines the basic properties of subjects. %
		% It is intended to have descendants, which between them %
		%     add frame rules and %
		%     control visibility of specification queries and %
		%     provide an implementation. "


deferred class SUBJECT_TYPE


feature -- basic query

	observers : IMMUTABLE_SET[ ABSTRACT_OBSERVER ] is
			-- The observers attached to this subject
		deferred
		end


feature -- attach and detach observers

	attached( o : ABSTRACT_OBSERVER ) : BOOLEAN is
			-- Is 'o' one of this subject's
			-- attached observers?
		deferred
		ensure
			consistent_with_observers: 
				Result = observers.has( o )
		end
			

	attach( o : ABSTRACT_OBSERVER ) is
			-- Add 'o' to the set of objects that 
			-- will be informed if this subject changes. 
		require
			observer_exists:	
				o /= Void
			observer_not_attached:
				not attached( o )
		deferred
		ensure
			attached:
				( ( o /= Void) and (not attached(o)) ) implies
					attached( o )
		end

	detach( o : ABSTRACT_OBSERVER ) is
			-- Remove 'o' from the set of objects that 
			-- will be informed if this subject changes. 
		require
			observer_exists:	
				o /= Void
			observer_attached:
				attached( o )
		deferred
		ensure
			detached:
				( ( o/= Void ) and (attached(o)) ) implies
					not attached( o )
		end

feature -- initialization

	initialize is
			-- Initialize this subject to have no observers
		deferred
		ensure
			no_observers:
				observers.count = 0
		end


feature -- send notifications

	notify is
			-- Inform attached observers that this subject has 
			-- changed by calling 'abstract_update' on each one.
			--
			-- 'notify' should be called by any 
			-- subclass feature that changes the state of
			-- 'Current' in a way that its observers should 
			-- know about.
		deferred
		ensure
			observers_up_to_date:
				all_observers_up_to_date( observers )
		end

	all_observers_up_to_date( current_observers : like observers ) : BOOLEAN is
		deferred
		ensure
			Result = 
				(
					( current_observers.count = 0 )
				  or else
					(
						current_observers.choice.up_to_date_with_subject
					  and
						all_observers_up_to_date( 
								current_observers.minus( current_observers.choice ) )
					)
				)
	end

end -- class SUBJECT_TYPE

