indexing
	description:
		" A decrement button has an associated counter. When %
		% the mouse is clicked over the screen view of the %
		% button, the button receives a call to the %
		% 'on_bn_clicked' routine, and decrements the counter. %
		% %
		% A decrement button is an observer of its associated counter. %
		% When the counter changes, the button receives a call %
		% to 'on_subject_changed', which disables the button if the %
		% associated counter is zero and enables the button if the %
		% associated counter is positive. This ensures that %
		% 'on_bn_clicked' is never called when the associated %
		% counter is zero. This, in turn, ensures that this object %
		% never tries to decrement its associated counter when %
		% the counter's value is zero. "


class DECREMENT_BUTTON

inherit
	WEL_PUSH_BUTTON
			-- A decrement button is a WEL push button
		rename
			make as base_initialize
		export {NONE} all {ANY} enabled
			-- Leave only the query 'enabled' visible, for
			-- use in contracts. The query 'enabled'
			-- is a basic query.
		redefine
			on_bn_clicked
				-- It has an empty body in the parent class.
		end

	OBSERVER
			-- A decrement button is an observer of a counter

create
	initialize

feature -- Basic query: the button's associated counter

	counter : COUNTER

feature -- Creation command

	initialize( a_counter : COUNTER;
		a_parent: WEL_COMPOSITE_WINDOW;
		a_name: STRING;				-- text on button
		an_x, a_y : INTEGER;			-- location of button
		a_width, a_height : INTEGER;		-- size of button
		an_id: INTEGER ) is

			-- Make a button that can decrement the given counter.
			-- Register it as an observer of the counter.
			-- Enable or disable it depending on the counter's state.
		do
			base_initialize( a_parent, a_name, an_x, a_y, a_width, a_height, an_id )
			counter := a_counter
			counter.register( Current )
			set_enabled_state_to_match_counter_state
		ensure
			counter_set:
				counter = a_counter
			enabled_status_consistent_with_counter_value:
				enabled = (counter.value > 0)
		end

feature -- Other commands: callbacks

	on_bn_clicked is
			-- (Called by the WEL event loop, provided the
			-- button is enabled.)
			--
			-- Call 'perform_decrement' to decrement
			-- the associated counter.
		do
			perform_decrement
		end


	on_subject_changed( a_subject : SUBJECT ) is
			-- (Called by this observers's subject
			-- whenever the subject changes.)
			--
			-- If the counter is zero, disable myself.
			-- If the counter is non-zero, enable myself.
		do
			set_enabled_state_to_match_counter_state
		ensure then
			enabled_status_consistent_with_counter_value:
				enabled = (counter.value > 0)
		end

feature {NONE} -- Subroutines

	perform_decrement is
			-- (subroutine of 'on_bn_clicked')
			-- Actually perform the decrement for 'on_bn_clicked'
			--
			-- Need a subroutine so we can have the precondition
			-- (can't put it on 'on_bn_clicked' because it would
			-- be taken as a strengethening of an inherited
			-- precondition).
		require
			enabled:
				enabled
		do
			check
				enabled_status_consistent_with_counter_value:
					enabled = (counter.value > 0)
			end
			counter.decrement
		ensure then
			value_of_associated_counter_decremented:
				counter.value = old counter.value - 1
		end


	set_enabled_state_to_match_counter_state is
			-- (subroutine of 'initialize' and 'on_subject_changed')
			-- If the counter is zero, disable myself.
			-- If the counter is non-zero, enable myself.
		do
			if
				counter.value=0
			then
				disable
			else
				enable
			end
		ensure
			enabled_status_consistent_with_counter_value:
				enabled = (counter.value > 0)
		end


end -- class DECREMENT_BUTTON
